# -*- coding: utf-8 -*-
#
# Copyright 2020 Osia authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Module exposes configuration to setup DNS needed for Openshift to work"""
from osia.installer.dns.nsupdate import NSUpdate
from osia.installer.dns.route53 import Route53Provider
from osia.installer.dns.base import DNSProvider

DNSProvider.register_provider('nsupdate', NSUpdate)
DNSProvider.register_provider('route53', Route53Provider)
